unit datau;

// Copyright  2002 by Ziff Davis Media, Inc.
// Written by Neil J. Rubenking

interface

uses
  Windows, SysUtils, Classes, DB, DBClient;

type
  Tdm1 = class(TDataModule)
    cdsTwks: TClientDataSet;
    cdsData : TClientDataSet;
    cdsUndo : TClientDataSet;
    cdsNote: TClientDataSet;
    dsTwks: TDataSource;
    dsData  : TDataSource;
    dsUndo  : TDataSource;
    dsNote: TDataSource;
    procedure DataModuleCreate(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    Status : Integer;
  end;

var
  dm1: Tdm1;

implementation
uses midconst;

{$R *.dfm}

procedure Tdm1.DataModuleCreate(Sender: TObject);
const
  badMsg = 'The database of %s (%s) is %s. RegistryRobot will create a '+
    'new empty %1:s.%3:s'#13#10#13#10'If you can locate the old %1:s '+
    'file, close RegistryRobot and copy the old file over the '+
    'newly-created empty one.';
VAR S : String;

  procedure cdsUndoRebuildDefs;
  const names : ARRAY[0..8] OF String = ('TwkID', 'Root', 'Key',
    'Value', 'DataType', 'ChangeType', 'Special', 'Data', 'VerMask');
  const types : ARRAY[0..8] OF TFieldType = (ftInteger, ftString,
    ftString, ftString, ftInteger, ftInteger, ftString, ftString,
    ftSmallInt);
  const sizes : ARRAY[0..8] OF Integer = (0, 4, 255, 255, 0, 0, 255, 255, 0);
  VAR N : Integer;
  begin
    cdsUndo.FieldDefs.Clear;
    FOR N := 0 TO 8 DO
      WITH cdsUndo.FieldDefs.AddFieldDef DO
        begin
          Name := Names[N];
          Attributes := [faUnNamed];
          DataType := types[N];
          Size := sizes[N];
        end;
  end;

  procedure cdsNoteRebuildDefs;
  const names : ARRAY[0..1] OF String = ('TwkID', 'Note');
  const types : ARRAY[0..1] OF TFieldType = (ftInteger, ftString);
  const sizes : ARRAY[0..1] OF Integer = (0, 4096);
  VAR N : Integer;
  begin
    cdsNote.FieldDefs.Clear;
    FOR N := 0 TO 1 DO
      WITH cdsNote.FieldDefs.AddFieldDef DO
        begin
          Name := Names[N];
          Attributes := [faUnNamed];
          DataType := types[N];
          Size := sizes[N];
        end;
  end;

begin
  Status := 0;
  // We use ParamStr(0) rather than Application.Exename because the
  // data module doesn't have access to the Application variable
  S := ExtractFileDir(ParamStr(0));
  cdsTwks.FileName := S + '\Twks.xml';
  cdsData.FileName := S + '\Data.xml';
  cdsUndo.FileName := S + '\Undo.xml';
  cdsNote.FileName := S + '\Note.xml';
  //NOTE: The main .dpr file contains a test for the presence of the
  //  essential twks.xml and Data.xml files. If they are not present,
  //  it never gets to this point. However, if they are corrupted a
  //  status flag gets set and the main window handles it. (We can't
  //  call Application.Terminate from this module).
  try
    cdsTwks.Open;
  except
    On E:Exception DO
      begin
        IF AnsiCompareText(E.Message, SErrorLoadingMidas) =0 THEN
          Status := 3
        ELSE
          Status := 1;
        Exit;
      end;
  end;
  try
    cdsData.Open;
  except
    On Exception DO
      begin
        Status := 2;
        Exit;
      end;
  end;
  IF FileExists(cdsUndo.FileName) THEN
    begin
      try
        cdsUndo.Open;
      except
        ON Exception DO
          begin
            MessageBox(0,PChar(Format(badMsg, ['stored undo actions',
              'undo.xml', 'corrupted', ' You will not be able to '+
              'undo tweaks that were used previously.'])),
              'RegistryRobot', MB_OK OR MB_ICONINFORMATION);
            // IF the file was a valid XML file with the wrong schema
            // then the field definitions are screwed up at this point.
            cdsUndoRebuildDefs;
            cdsUndo.CreateDataSet;
            cdsUndo.SaveToFile(cdsUndo.FIlename);
            cdsUndo.Open;
          end;
      end;
    end
  ELSE
    begin
      MessageBox(0,PChar(Format(badMsg, ['stored undo actions',
        'undo.xml', 'missing', ' You will not be able to '+
        'undo tweaks that were used previously.'])),
        'RegistryRobot', MB_OK OR MB_ICONINFORMATION);
      cdsUndo.CreateDataSet;
      cdsUndo.SaveToFile(cdsUndo.FIlename);
      cdsUndo.Open;
    end;
  IF FileExists(cdsNote.FileName) THEN
    try
      cdsNote.Open;
    except
      ON Exception DO
        begin
          MessageBox(0,PChar(Format(badMsg, ['your notes',
            'note.xml', 'corrupted', ''])), 'RegistryRobot',
            MB_OK OR MB_ICONINFORMATION);
          // IF the file was a valid XML file with the wrong schema
          // then the field definitions are screwed up at this point.
          cdsNoteRebuildDefs;
          cdsNote.CreateDataSet;
          cdsNote.SaveToFile(cdsNote.FIlename);
          cdsNote.Open;
        end;
    end
  ELSE
    begin
      MessageBox(0,PChar(Format(badMsg, ['your notes', 'note.xml',
        'missing', ''])), 'RegistryRobot', MB_OK OR
        MB_ICONINFORMATION);
      cdsNote.CreateDataSet;
      cdsNote.SaveToFile(cdsNote.FIlename);
      cdsNote.Open;
    end;
  cdsTwks.LogChanges := False;
  cdsData.LogChanges := False;
  cdsUndo.LogChanges := False;
  cdsNote.LogChanges := False;
end;

end.


