unit rrfuncs;

// Copyright  2002 by Ziff Davis Media, Inc.
// Written by Neil J. Rubenking

interface

uses Windows, SysUtils, Forms, Controls, ImgList;

function CurrentPlatform : Integer;
function IEVersion : Integer;
function GetVerMask(VAR S : String) : Word;
function PlatString(vMask : Word): String;
function SpecialNum(const S : String) : Integer;
function RootFor(const S : String) : hKey;
function GetFolder(F : TForm; const S, Title : String;
  VAR GotS : String) : Boolean;
function FirstWord(VAR S : String) : String;
function Shorten(const S : String) : String;
procedure MyExpand(VAR S : String);
procedure LoadImageList(IL : TImageList; iFrom, iTo: Integer);
function GetPrivMask : Integer;
function KeyNames(mask : Integer) : String;

const
  w95 = 0;
  w98 = 1;
  wME = 2;
  wNT = 3; //NT4
  w20 = 4; //2000
  wXP = 5;

  pHKCU = $01;
  pHKCR = $02;
  pHKCC = $04;
  pHKLM = $08;
  pHKU  = $10;

  c95 = $0001;
  c98 = $0002;
  cME = $0004;
  c9x = c95 OR c98 OR cME;
  cNT = $0008;
  c20 = $0010;
  cXP = $0020;
  cNTx = cNT OR c20 OR cXP;
  cI4 = $0100;
  cI5 = $0200;
  cI6 = $0400;
  cWD = $0800; //WDU
  cI4p = cI4 OR cI5 OR cI6;
  cI5p = cI5 OR cI6;

  specnone    = 0;
  specdrive   = 1;
  specfile    = 2;
  specfolder  = 3;
  specicon    = 4;
  specmemo    = 5;
  specnum     = 6;
  spectext    = 7;
  spectextb   = 8;
  specview    = 9;
  spectoggle  = 10;
  speckeyvar  = 11;
  specexpand  = 12;


  valtypenames : ARRAY[0..4] OF String =
    ('(none)', 'String', 'Expand', 'Binary', 'DWORD');
  ctAddKey = 0;
  ctChgVal = 1;
  ctDelKey = 2;
  ctDelVal = 3;
  ctVueVal = 4;
  chgtypenames : ARRAY[0..4] OF String =
    ('Add key', 'Change value', 'Delete key', 'Delete value',
    'View value');

VAR ShellVersion : DWORD; 

implementation
uses classes, shlobj, ShellApi, ActiveX, comobj, graphics,
  Registry, FileCtrl;

VAR
  windir, sysdir : String;

function CurrentPlatform : Integer;
begin
  Result := -1;
  IF Win32Platform = VER_PLATFORM_WIN32_WINDOWS THEN
    begin
      IF (Win32MajorVersion = 4) AND (Win32MinorVersion=0) THEN
        Result := w95
      ELSE IF (Win32MajorVersion = 4) AND (Win32MinorVersion=10) THEN
        Result := w98
      ELSE IF (Win32MajorVersion = 4) AND (Win32MinorVersion=90) THEN
        Result := wME;
    end
  ELSE IF Win32Platform = VER_PLATFORM_WIN32_NT THEN
    begin
      IF Win32MajorVersion = 4 THEN
        Result := wNT
      ELSE IF (Win32MajorVersion = 5) AND (Win32MinorVersion=0) THEN
        Result := w20
      ELSE IF (Win32MajorVersion = 5) AND (Win32MinorVersion=1) THEN
        Result := wXP;
    end;
end;

function IEVersion : Integer;
// Reports versions from 4 on - pre-4 is nothing
VAR S : String;
begin
  Result := -1;
  WITH TRegistry.Create DO
  try
    RootKey := HKEY_LOCAL_MACHINE;
    IF OpenKeyReadOnly('Software\Microsoft\Internet Explorer') THEN
      begin
        IF ValueExists('Version') THEN
          begin
            S := ReadString('Version');
            IF S <> '' THEN
              Result := Ord(S[1])-Ord('0');
          end;
      end;
  finally
    Free;
  end;
end;

function GetVerMask(VAR S : String) : Word;
  function HasWDU : Boolean;
  begin
    Result := FindWindow('BaseBar', nil) <> 0;
  end;
begin
  S := '';
  Result := 0;
  CASE CurrentPlatform OF
    w95 : IF HasWDU THEN
      begin
        Result := Result OR c95 OR cWD;
        S := '95+WDU';
      end
    ELSE
      begin
        Result := Result OR c95;
        S := '95';
      end;
    w98 : begin
      Result := Result OR c98 OR cWD;
      S := '98';
    end;
    wME : begin
      Result := Result OR cME OR cWD;
      S := 'ME';
    end;
    wNT : IF HasWDU THEN
      begin
        Result := Result OR cNT OR cWD;
        S := 'NT+WDU';
      end
    ELSE
      begin
        Result := Result OR cNT;
        S := 'NT';
      end;
    w20 : begin
      Result := Result OR c20 OR cWD;
      S := '2000';
    end;
    wXP : begin
      Result := Result OR cXP OR cWD;
      S := 'XP';
    end;
  END;
  CASE IEVersion OF
    4 : begin
      Result := Result OR cI4;
      S := S + ', IE4';
    end;
    5 : begin
      Result := Result OR cI5;
      S := S + ', IE5';
    end;
    6 : begin
      Result := Result OR cI6;
      S := S + ', IE6';
    end;
  END;
end;

function PlatString(vMask : Word) : String;
const
  Platnames : ARRAY[0..5] OF String = ('95, ', '98, ',
    'ME, ', 'NT4, ', '2000, ', 'XP, ');
begin
  Result := '';
  IF vMask AND c95 > 0 THEN
      Result := Result + '95, '
  ELSE IF vMask AND cWD > 0 THEN
      Result := Result + '95+WDU, ';
  IF vMask AND c98 > 0 THEN Result := Result + '98, ';
  IF vMask AND cME > 0 THEN Result := Result + 'ME, ';
  IF vMask AND cNT > 0 THEN
      Result := Result + 'NT4, '
  ELSE IF vMask AND cWD > 0 THEN
      Result := Result + 'NT4+WDU, ';
  IF vMask AND c20 > 0 THEN Result := Result + '2000, ';
  IF vMask AND cXP > 0 THEN Result := Result + 'XP, ';
  IF Result <> '' THEN SetLength(Result, Length(Result)-2);
  IF vMask AND cI4p = 0 THEN Exit;
  IF Result <> '' THEN result := Result + '; ';
  CASE vMask AND cI4p OF
    0 : Exit;
    cI4p : Result := Result + 'IE4+';
    cI5p : Result := Result + 'IE5+';
    ELSE
      begin
        IF Result <> '' THEN
          Result := Result + '; ';
        IF vMask AND cI4 > 0 THEN Result := Result + 'IE4, ';
        IF vMask AND cI5 > 0 THEN Result := Result + 'IE5, ';
        IF vMask AND cI6 > 0 THEN Result := Result + 'IE6, ';
        SetLength(Result, Length(Result)-2);
      end;
  end;
end;

function SpecialNum(const S : String) : Integer;
const
  specials : ARRAY[specDrive..specexpand] OF String =
    ('drive', 'file', 'folder', 'icon', 'memo', 'num',
     'text', 'textb', 'view', 'toggle', 'readkeyvar','expand');
VAR N : Integer;
begin
  Result := specNone;
  FOR N := specDrive TO specexpand DO
    IF AnsiCompareText(S, specials[N]) = 0 THEN
      begin
        Result := N;
        Break;
      end;
end;

function RootFor(const S : String) : hKey;
const roots = '^HKCR^HKCU^HKLM^HKU^HKCC^';
begin        //12345678901234567890
  CASE Pos(Format('^%s^', [S]), Roots) OF
    01 : Result := HKEY_CLASSES_ROOT;
    06 : Result := HKEY_CURRENT_USER;
    11 : Result := HKEY_LOCAL_MACHINE;
    16 : Result := HKEY_USERS;
    20 : Result := HKEY_CURRENT_CONFIG;
    ELSE Result := 0; // Deal with error
  END;
end;

type
  PPassData = ^TPassData;
  TPassData = record
    buff : ARRAY[0..MAX_PATH] OF Char;
    Wid  : Integer;
    Can  : TCanvas;
  end;

function BrowseCallBackProc(hWindow : HWND; uMsg : Integer;
  lParam : LPARAM; lpData : PPassData) : Integer; stdcall;
VAR
  buffer : ARRAY[0..MAX_PATH] OF Char;
  Dummy : DWord;
  SF : SHFileInfo;
  R : TRect;
  SDisp : String;
begin
  Result := 0;
  IF uMsg = BFFM_INITIALIZED THEN
    begin
      GetClientRect(hWindow, R);
      lpData.Wid := R.Right - R.Left - 20;
      StrCopy(Buffer, lpData.buff);
      StrPCopy(Buffer, ExcludeTrailingBackslash(StrPas(Buffer)));
      // Set the starting directory
      IF (StrLen(Buffer) = 2) AND (Buffer[1] = ':') THEN
        StrCat(Buffer, '\');
      SendMessage(hWindow, BFFM_SETSELECTION, 1, Integer(@Buffer));
    end
  ELSE IF uMsg = BFFM_SELCHANGED THEN
    begin
      FillChar(SF, SizeOf(SF), 0);
      SHGetFileInfo(PChar(lParam), Dummy, SF,
        SizeOf(SF), SHGFI_PIDL	OR SHGFI_TYPENAME);
      IF AnsiCompareText(SF.szTypeName, 'Folder Shortcut') = 0 THEN
        begin
          SendMessage(hWindow, BFFM_ENABLEOK, 0, 0);
          StrCopy(Buffer, 'Folder shortcut, not a valid selection '+
            '- see Help.');
          SendMessage(hWindow, BFFM_SETSTATUSTEXT, 0,
            Integer(@Buffer));
        end
      ELSE
        begin
          SHGetPathFromIDList(pItemIDList(lParam), buffer);
          IF Buffer[0] = #0 THEN
            SendMessage(hWindow, BFFM_ENABLEOK, 0, 0);
          // Display the current selection
          SDisp := MinimizeName(StrPas(Buffer), lpData.Can,
            lpData.Wid);
          StrPCopy(Buffer, SDisp);
          SendMessage(hWindow, BFFM_SETSTATUSTEXT, 0,
            Integer(@Buffer));
        end;
    end
  ELSE
    begin
      StrPCopy(Buffer, Format('"%s" is not a valid folder name',
        [PChar(lParam)]));
      SendMessage(hWindow, BFFM_SETSTATUSTEXT, 0, Integer(@Buffer));
      MessageBeep(MB_ICONSTOP);
      Result := 1;
    end;
end;

function GetFolder(F : TForm; const S, Title : String;
  VAR GotS : String) : Boolean;
const BIF_NEWDIALOGSTYLE = $40;
VAR
  TB     : TBrowseInfo;
  IDL    : PItemIdList;
  buffer : ARRAY[0..MAX_PATH] OF Char;
  TPass  : TPassData;
begin
  FillChar(TB, SizeOf(TB), 0);
  // Connect it to our window
  TB.hwndOwner := F.Handle;
  // Set the browse-dialog's title
  TB.lpszTitle := PChar(Title);
  TB.ulFlags   := BIF_RETURNONLYFSDIRS OR BIF_RETURNFSANCESTORS
    OR BIF_STATUSTEXT ;
  // If the shell supports it, include an edit box
  IF ShellVersion >= 471 THEN
    TB.ulFlags := TB.ulFlags OR BIF_EDITBOX OR BIF_VALIDATE;
  // Set the browse callback procedure
  TB.lpfn      := @BrowseCallBackProc;
  // point the lParam at the desired starting folder
  // point the lParam at the desired starting folder
  StrPCopy(TPass.buff, S);
  TPass.Can := F.Canvas;
  TPass.Wid := 0;
  TB.lParam    := Integer(@TPass);
  IDL := ShBrowseForFolder(TB);
  IF IDL <> nil THEN
    begin
      SHGetPathFromIDList(IDL, Buffer);
      CoTaskMemFree(IDL);
      GotS   := IncludeTrailingBackSlash(StrPas(Buffer));
      Result := True;
    end
  ELSE Result := False;
end;

function FirstWord(VAR S : String) : String;
VAR P : Integer;
begin
  Result := S;
  P := Pos('|', S);
  IF P = 0 THEN S := ''
  ELSE
    begin
      SetLength(Result, P-1);
      System.Delete(S, 1, P);
    end;
end;

function Shorten(const S : String) : String;
begin
  Result := S;
  IF Length(Result) > 60 THEN
    Result := Copy(Result, 1, 60) + '...';
end;

function GetShellVersion : DWORD;
// 4.70 = 470, 5.0 = 500, etc. ANything before 4.70 = 0
type
  PDllVersionInfo = ^TDllVersionInfo;
  TDllVersionInfo = record
    cbSize          : DWORD;
    dwMajorVersion  : DWORD;
    dwMinorVersion  : DWORD;
    dwBuildNumber   : DWORD;
    dwPlatformID    : DWORD;
  end;
  DllGetVersionType = function (pdvi : PDllVersionInfo) :
    HRESULT; stdcall;
VAR
  LibH : hModule;
  DllGetVersion  : DllGetVersionType;
  DVI  : TDllVersionInfo;
begin
  Result := 0;
  LibH := LoadLibrary('Shell32.dll');
  IF LibH <= 32 THEN Exit;
  try
    @DllGetVersion := GetProcAddress(LibH, 'DllGetVersion');
    IF @DllGetVersion = nil THEN Exit;
    FillChar(DVI, SizeOf(DVI), 0);
    DVI.cbSize := SizeOf(DVI);
    IF DllGetVersion(@DVI) = NOERROR THEN
      begin
        Result := DVI.dwMajorVersion * 100;
        IF DVI.dwMinorVersion > 100 THEN Result := Result + 99
        ELSE Result := Result + DVI.dwMinorVersion;
      end;
  finally
    FreeLibrary(LibH);
  end;
end;

procedure MyExpand(VAR S : String);
begin
  S := StringReplace(S, '%windir%', windir,
    [rfReplaceAll, rfIgnoreCase]);
  S := StringReplace(S, '%sysdir%', sysdir,
    [rfReplaceAll, rfIgnoreCase]);
end;

procedure LoadImageList(IL : TImageList; iFrom, iTo: Integer);
// Load the image list with resource bitmaps in the specified
// range. By doing so, we save space and avoid complications
// with earlier versions of COMCTL32.DLL
VAR
  TB       : TBitmap;
  N, W, H  : Integer;
begin
  TB := TBitmap.Create;
  try
    IL.Clear;
    W := IL.Width;
    H := IL.Height;
    FOR N := iFrom TO iTo DO
      begin
        TB.Handle := LoadImage(hInstance, MakeIntResource(N),
          IMAGE_BITMAP, W, H, lr_DefaultColor);
        IL.AddMasked(TB, clFuchsia);
      end;
  finally
    TB.Free;
  end;
end;

function CanWrite(H : hKey) : Boolean;
VAR Gs : String;
  function GetGuid : TGuid;
  begin
    OleCheck(CoCreateGuid(Result));
  end;
begin
  WITH TRegistry.Create DO
  try
    RootKey := H;
    REPEAT
      Gs := GuidToString(GetGuid);
      CASE H OF
        HKEY_LOCAL_MACHINE : Gs := 'SOFTWARE\'+Gs;
        HKEY_USERS         : Gs := '.Default\'+Gs;
      END;
    UNTIL NOT KeyExists(Gs);
    try
      CreateKey(Gs);
      DeleteKey(Gs);
      Result := True;
    except
      ON Exception DO
        Result := False;
    end;
  finally
    Free;
  end;
end;

function GetPrivMask : Integer;
begin
  Result := 0;
  IF CanWrite(HKEY_CURRENT_USER)   THEN Result := Result OR pHKCU;
  IF CanWrite(HKEY_CLASSES_ROOT)   THEN Result := Result OR pHKCR;
  IF CanWrite(HKEY_CURRENT_CONFIG) THEN Result := Result OR pHKCC;
  IF CanWrite(HKEY_LOCAL_MACHINE)  THEN Result := Result OR pHKLM;
  IF CanWrite(HKEY_USERS)          THEN Result := Result OR pHKU;
end;

function KeyNames(mask : Integer) : String;
VAR TS : TStringList;
begin
  TS := TStringList.Create;
  try
    IF mask AND pHKCU > 0 THEN TS.Add('HKEY_CURRENT_USER');
    IF mask AND pHKCR > 0 THEN TS.Add('HKEY_CLASSES_ROOT');
    IF mask AND pHKCC > 0 THEN TS.Add('HKEY_CURRENT_CONFIG');
    IF mask AND pHKLM > 0 THEN TS.Add('HKEY_LOCAL_MACHINE');
    IF mask AND pHKU  > 0 THEN TS.Add('HKEY_USERS');
    IF TS.Count > 1 THEN
      TS.Insert(TS.Count-1, 'and');
    IF TS.Count = 3 THEN
      Result := StringReplace(TS.CommaText,',', ' ', [rfReplaceAll])
    ELSE
    Result := StringReplace(
      StringReplace(TS.CommaText,',', ', ', [rfReplaceAll]),
      'and,', 'and', []);
  finally
    TS.Free;
  end;
end;

initialization
  SetLength(windir, MAX_PATH);
  SetLength(windir, GetWindowsDirectory(pChar(winDir), MAX_PATH));
  SetLength(sysdir, MAX_PATH);
  SetLength(sysdir, GetSystemDirectory(pChar(sysDir), MAX_PATH));
  windir := ExcludeTrailingBackslash(windir);
  sysdir := ExcludeTrailingBackslash(sysdir);
  ShellVersion := GetShellVersion;
  IF ShellVersion >= 500 THEN
    CoInitialize(nil);
finalization
  IF ShellVersion >= 500 THEN
    CoUninitialize;
end.
